# Mathematical expression analyser for MaExPa.
#
# Copyright 2020 Alexandre Emsenhuber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import exception, tokens

class Expression( object ):
	def __init__( self, expr, var = None, func = None ):
		self.expr = expr
		self.tokens = tokens.lexer( expr )
		self.var_cb = var
		self.func_cb = func
		self.cur = None
		self.value = None

	def next_val( self, s ):
		if self.cur < len( self.tokens ) and self.tokens[ self.cur ].get_type() == s:
			self.value = self.tokens[ self.cur ].get_value()
			self.cur += 1
			return True
		else:
			return False

	def next_is( self, s ):
		if self.cur < len( self.tokens ) and self.tokens[ self.cur ].get_type() == s:
			self.value = None
			self.cur += 1
			return True
		else:
			return False

	def next_must( self, s ):
		if not self.next_is( s ):
			self.error()

	def operator( self, allowed ):
		if self.cur < len( self.tokens ) and self.tokens[ self.cur ].get_type() == "operator":
			if self.tokens[ self.cur ].get_value() in allowed:
				self.value = self.tokens[ self.cur ]
				self.cur += 1
				return True
		return False

	def error( self ):
		if self.cur >= len( self.tokens ):
			raise exception.ParseEndException( None, -1 )
		else:
			raise exception.ParseException( self.tokens[ self.cur ].get_type(), self.tokens[ self.cur ].get_column() + 1 )

	def __call__( self, var = None, func = None ):
		self.cur = 0
		self.value = None

		if not var is None:
			var_save = self.var_cb
			self.var_cb = var
		if not func is None:
			func_save = self.func_cb
			self.func_cb = func

		try:
			value = self.expression()
		finally:
			if not var is None:
				self.var_cb = var_save
			if not func is None:
				self.func_cb = func_save

		if self.cur != len( self.tokens ):
			self.error()

		return value

	def expression( self ):
		self.operator( [ "+", "-" ] )
		oper = self.value
		value = self.term()
		if oper:
			value = oper.unary( value )

		while self.operator( [ "+", "-" ] ):
			oper = self.value
			value = oper.binary( value, self.term() )

		return value

	def term( self ):
		value = self.factor()

		while self.operator( [ "*", "/", "//" ] ):
			oper = self.value
			value = oper.binary( value, self.factor() )

		return value

	def factor( self ):
		# This one is right associative :(
		bases = [ self.base() ]
		while self.operator( [ "**" ] ):
			bases.append( self.base() )

		ret = bases[ -1 ]
		for i in range( 1, len( bases ) ):
			ret = bases[ len( bases ) - i - 1 ] ** ret

		return ret

	def base( self ):
		if self.next_val( "name" ):
			name = self.value
			if self.next_is( "(" ):
				args = [ self.expression() ]
				while self.next_is( "," ):
					args.append( self.expression() )
				self.next_must( ")" )
				return self.function( name, args )
			else:
				return self.variable( name )

		if self.next_val( "float" ):
			return self.value

		if self.next_val( "integer" ):
			return self.value

		if self.next_is( "(" ):
			value = self.expression()
			self.next_must( ")" )
			return value

		self.error()

	def function( self, name, args ):
		if self.func_cb is None:
			raise exception.NoFuncException( name, -1 )

		return self.func_cb( name, args )

	def variable( self, name ):
		if self.var_cb is None:
			raise exception.NoVarException( name, -1 )

		return self.var_cb( name )



# Lexical tokens
#
# Copyright 2020 Alexandre Emsenhuber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import operator

from . import exception

_re_int = re.compile( "[0-9]+" )
_re_float = re.compile( "([0-9]*)([.][0-9]*)?([eE][+-]?[0-9]+)?" )
_re_name = re.compile( "[a-zA-Z_][a-zA-Z_0-9]*" )

class TokenBase( object ):
	"""
	Base class for lexical tokens.
	"""

	def __init__( self, col ):
		self.col = col

	def get_column( self ):
		"""
		Get the column (or character number) of the beginning of this token.
		"""
		return self.col

	def get_type( self ):
		"""
		Get the type of this token as a str.
		"""
		raise NotImplementedError

	def get_value( self ):
		"""
		Get the
		"""
		raise NotImplementedError

class TokenName( TokenBase ):
	"""
	Token representing a variable or function reference.
	"""

	def __init__( self, col, value ):
		TokenBase.__init__( self, col )
		self.value = value

	def get_type( self ):
		return "name"

	def get_value( self ):
		return self.value

class TokenInt( TokenBase ):
	"""
	Token representing a litteral integer.
	"""

	def __init__( self, col, value ):
		TokenBase.__init__( self, col )
		self.value = value

	def get_type( self ):
		return "integer"

	def get_value( self ):
		return int( self.value )

class TokenFloat( TokenBase ):
	"""
	Token representing a litteral floating-point number.
	"""

	def __init__( self, col, value ):
		TokenBase.__init__( self, col )
		self.value = value

	def get_type( self ):
		return "float"

	def get_value( self ):
		return float( self.value )

class TokenOperator( TokenBase ):
	"""
	Token representing a mathematical operator.
	"""

	_map = {
		"**": ( operator.pow, ),
		"//": ( operator.floordiv, ),
		"*": ( operator.mul, ),
		"/": ( operator.truediv, ),
		"+": ( operator.add, operator.pos ),
		"-": ( operator.sub, operator.neg ),
	}

	def __init__( self, col, value ):
		TokenBase.__init__( self, col )
		self.value = value

	def get_type( self ):
		return "operator"

	def get_value( self ):
		return self.value

	def unary( self, value ):
		if len( self._map[ self.value ] ) == 1:
			raise Exception( "Operator {:s} is not unary".format( self.value ) )

		return self._map[ self.value ][ 1 ]( value )

	def binary( self, left, right ):
		return self._map[ self.value ][ 0 ]( left, right )

class TokenLPar( TokenBase ):
	"""
	Token representing a litteral "(".
	"""

	def get_type( self ):
		return "("

class TokenRPar( TokenBase ):
	"""
	Token representing a litteral ")".
	"""

	def get_type( self ):
		return ")"

class TokenComma( TokenBase ):
	"""
	Token representing a litteral ",".
	"""

	def get_type( self ):
		return ","

def lexer( formula ):
	"""
	Lexical analysis.
	"""

	start = 0
	tot = len( formula )

	tklist = []

	while start < tot:
		if formula[ start ] == " ":
			start += 1
			continue

		match = _re_float.match( formula, start )
		if match:
			if ( not match.group( 2 ) is None and len( match.group( 2 ) ) > 1 ) or ( len( match.group( 1 ) ) > 0 and ( not match.group( 2 ) is None or not match.group( 3 ) is None ) ):
				tklist.append( TokenFloat( start, match.group() ) )
				_, start = match.span()
				continue

		match = _re_int.match( formula, start )
		if match:
			tklist.append( TokenInt( start, match.group() ) )
			_, start = match.span()
			continue

		match = _re_name.match( formula, start )
		if match:
			tklist.append( TokenName( start, match.group() ) )
			_, start = match.span()
			continue

		for op in [ "**", "//", "*", "/", "+", "-" ]:
			if formula[ start : start+len( op ) ] == op:
				tklist.append( TokenOperator( start, op ) )
				start += len( op )
				break
		else:
			if formula[ start ] == "(":
				tklist.append( TokenLPar( start ) )
				start += 1
			elif formula[ start ] == ")":
				tklist.append( TokenRPar( start ) )
				start += 1
			elif formula[ start ] == ",":
				tklist.append( TokenComma( start ) )
				start += 1
			else:
				raise exception.LexicalException( formula[ start ], start + 1 )

	return tklist



# Exceptions for MaExPa.
#
# Copyright 2020 Alexandre Emsenhuber
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class CommonException( Exception ):
	def __init__( self, desc, column ):
		self.desc = desc
		self.column = column

	def __str__( self ):
		return "{:s} at column {:d}".format( self.desc, self.column )

class LexicalException( CommonException ):
	def __str__( self ):
		return "Parse Error: unexpected character `{:s}' at column {:d}".format( self.desc, self.column )

class ParseException( CommonException ):
	def __str__( self ):
		return "Parse Error: unexpected symbol `{:s}' at column {:d}".format( self.desc, self.column )

class ParseEndException( CommonException ):
	def __str__( self ):
		return "Parse Error: unexpected end"

class NoVarException( CommonException ):
	def __str__( self ):
		return "Runtime Error: variable `{:s}' does not exist".format( self.desc )

class NoFuncException( CommonException ):
	def __str__( self ):
		return "Runtime Error: function `{:s}' does not exist".format( self.desc )



import setuptools

setuptools.setup(
    name = "maexpa",
    version = "0.1",
    author = "Alexandre Emsenhuber",
    author_email = "emsenhuber@lpl.arizona.edu",
    description = "Extensible mathematical expression analyzer with callbacks for variables and functions",
    long_description = open( "README.md", "r" ).read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/aemsenhuber/maexpa",
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
